#include "stdafx.h"
#include "../../../code/vc6/alpha1/src/blepo.h"
#include "../LogTrace/Trace.h"
using namespace blepo;
#include "match_vnk.h"
#include "bnutility.h"

void SeeInt(int val)
{
  char buffer[100];
  sprintf(buffer,"value is %d",val);
  AfxMessageBox(buffer);
}

void SeeFloat(float val)
{
  char buffer[100];
  sprintf(buffer,"value is %f",val);
  AfxMessageBox(buffer);
}

void SeeLocation(Location p)
{
  char buffer[100];
  sprintf(buffer,"p.x %d p.y %d p.t %d",p.x,p.y,p.t);
  AfxMessageBox(buffer);
}

void WriteFloatVect(
  char* filename,
  const std::vector<float>& flt_vect)
{
  std::ofstream out_file(filename);
  char buffer[100];

  std::vector<float>::const_iterator i;
  for(i=flt_vect.begin(); i!=flt_vect.end(); i++)
  {
    sprintf(buffer,"%f ",*i);
    out_file << buffer;
  }

  out_file << "\n";
  out_file.close();
}

float MeanFloatVect(const std::vector<float>& flt_vect)
{
  std::vector<float>::const_iterator i;
  float sum=0.0f;
  for(i=flt_vect.begin(); i!=flt_vect.end(); i++)
    sum += (*i);

  float size = (float) flt_vect.size();

  if(size != 0.0f)
    return (sum/size);
  else
    return 0.0f;
}

int compare(const void * a, const void * b)
{
  if(*(int*)a < *(int*)b)
    return -1;
  else if(*(int*)a == *(int*)b)
    return 0;
  else
    return 1;
}

int comparereverse(const void * a, const void * b)
{
  if(*(int*)a < *(int*)b)
    return 1;
  else if(*(int*)a == *(int*)b)
    return 0;
  else
    return -1;
}

void MedianFilterFinal(const ImgInt& in, const int filter_width, const int filter_height, ImgInt* out)
{
  assert(in.Width() > 2);
  assert(in.Height() > 2);

  int wt = in.Width();
  int ht = in.Height();

  assert((filter_width>=3) && (filter_width%2!=0)); // filter width odd and atleast 3
  assert((filter_height>=3) && (filter_height%2!=0)); // filter height odd and atleast 3
  assert((filter_width<=wt) && (filter_height<=ht)); // filter size within image dimensions

  out->Reset(wt, ht);
  //Set(out, 0);

  int fw_hl = (filter_width-1)/2; // filter width half length
  int fh_hl = (filter_height-1)/2; // filter height half length

  int *in_data, itr, i, j, k, i_s, j_s, i_e, j_e;
  int box_length, box_half_length;

  Rect box;
  ImgInt::Iterator out_p = out->Begin();
  for(j=0; j<ht; j++)
    for(i=0; i<wt; i++)
    {
      i_s = i;
      for(k=0; k<fw_hl; k++)
      {
        if(i_s == 0) break;
        i_s--;
      }
      i_e = i;
      for(k=0; k<=fw_hl; k++)
      {
        i_e++;
        if(i_e == wt) break;
      }
      j_s = j;
      for(k=0; k<fh_hl; k++)
      {
        if(j_s == 0) break;
        j_s--;
      }
      j_e = j;
      for(k=0; k<=fh_hl; k++)
      {
        j_e++;
        if(j_e == ht) break;
      }

      box_length = (i_e-i_s)*(j_e-j_s);
      box_half_length = box_length/2;
      in_data = new int[box_length];
      box = Rect(i_s,j_s,i_e,j_e);
      ImgInt::ConstRectIterator box_p = in.BeginRect(box);
      itr = 0;
      while(!box_p.AtEnd())
      {
        *(in_data + itr) = *box_p;
        itr++;
        box_p++;
      }
      qsort(in_data, box_length, sizeof(int), compare);
      *out_p = *(in_data + box_half_length);
      out_p++;
      delete[] in_data;
    }
}

void WriteResults(
  const std::vector<MatFlt>& mat,
  const std::vector<float>& centroid_x,
  const std::vector<float>& centroid_y,
  const std::vector<int>& status,
  const char* filename)
{
  assert(mat.size() == centroid_x.size());
  assert(mat.size() == centroid_y.size());
  assert(mat.size() == status.size());

  std::ofstream result_file(filename);
  char buffer[100];

  std::vector<MatFlt>::const_iterator j;
  std::vector<float>::const_iterator x,y;
  std::vector<int>::const_iterator sts;
  x = centroid_x.begin();
  y = centroid_y.begin();
  sts = status.begin();
  int i=0;
  for(j=mat.begin(); j!=mat.end(); j++)
  {
    sprintf(buffer,"label is %d status is %d \t\t",i,*sts);
    result_file << buffer;
    for(int k=0; k<6; k++)
    {
      sprintf(buffer,"%f ",*(j->Begin(k,0)));
      result_file << buffer;
    }
    
    sprintf(buffer,"%f ",*x);
    result_file << buffer;
    sprintf(buffer,"%f ",*y);
    result_file << buffer;

    result_file << "\n";
    i++;
    x++;
    y++;
    sts++;
  }

  result_file.close();
}

void TraceDispFunc(
  const std::vector<MatFlt>& mat,
  const std::vector<float>& centroid_x,
  const std::vector<float>& centroid_y)
{
  assert(mat.size() == centroid_x.size());
  assert(mat.size() == centroid_y.size());

  char buffer[100];

  std::vector<MatFlt>::const_iterator j;
  std::vector<float>::const_iterator x,y;
  x = centroid_x.begin();
  y = centroid_y.begin();
  int i=0;
  for(j=mat.begin(); j!=mat.end(); j++)
  {
    sprintf(buffer,"label is %d \t\t",i);
    Trace(buffer);
    for(int k=0; k<6; k++)
    {
      sprintf(buffer,"%f ",*(j->Begin(k,0)));
      Trace(buffer);
    }
    
    sprintf(buffer,"%f ",*x);
    Trace(buffer);
    sprintf(buffer,"%f ",*y);
    Trace(buffer);

    Trace("\n");
    i++;
    x++;
    y++;
  }
}

void WriteResults(
  const std::vector<MatFlt>& mat,
  const std::vector<float>& centroid_x,
  const std::vector<float>& centroid_y,
  const std::vector<float>& residue,
  const std::vector<int>& status,
  const char* filename)
{
  assert(mat.size() == centroid_x.size());
  assert(mat.size() == centroid_y.size());
  assert(mat.size() == residue.size());
  assert(mat.size() == status.size());

  std::ofstream result_file(filename);
  char buffer[100];

  std::vector<MatFlt>::const_iterator j;
  std::vector<float>::const_iterator x,y,r;
  std::vector<int>::const_iterator sts;
  x = centroid_x.begin();
  y = centroid_y.begin();
  r = residue.begin();
  sts = status.begin();
  int i=0;
  for(j=mat.begin(); j!=mat.end(); j++)
  {
    sprintf(buffer,"label is %d status is %d residue is %f\t\t",i,*sts,*r);
    result_file << buffer;
    for(int k=0; k<6; k++)
    {
      sprintf(buffer,"%f ",*(j->Begin(k,0)));
      result_file << buffer;
    }
    
    sprintf(buffer,"%f ",*x);
    result_file << buffer;
    sprintf(buffer,"%f ",*y);
    result_file << buffer;

    result_file << "\n";
    i++;
    x++;
    y++;
    r++;
    sts++;
  }

  result_file.close();
}

void WriteResults(
  const std::vector<LKAResult>& result_r,
  const char* filename)
{
  std::ofstream result_file(filename);
  char buffer[100];

  std::vector<LKAResult>::const_iterator rr;
  int i=0;
  for(rr=result_r.begin(); rr!=result_r.end(); rr++)
  {
    sprintf(buffer,"label is %d status is %d residue is %f\t\t",i,rr->status,rr->residue);
    result_file << buffer;

    sprintf(buffer,"%f %f ",rr->A(0,0),rr->A(1,0));
    result_file << buffer;
    sprintf(buffer,"%f %f ",rr->d(0,0),rr->A(0,1));
    result_file << buffer;
    sprintf(buffer,"%f %f ",rr->A(1,1),rr->d(0,1));
    result_file << buffer;

    sprintf(buffer,"%f ",rr->centroid_x);
    result_file << buffer;
    sprintf(buffer,"%f ",rr->centroid_y);
    result_file << buffer;

    result_file << "\n";
    i++;
  }

  result_file.close();
}

void TraceLKAResultArray(const std::vector<LKAResult>& result_r)
{
  char buffer[100];

  std::vector<LKAResult>::const_iterator rr;
  int i=0;
  for(rr=result_r.begin(); rr!=result_r.end(); rr++)
  {
    sprintf(buffer,"label is %d status is %d residue is %f\t\t",i,rr->status,rr->residue);
    Trace(buffer);

    sprintf(buffer,"%f %f ",rr->A(0,0),rr->A(1,0));
    Trace(buffer);
    sprintf(buffer,"%f %f ",rr->d(0,0),rr->A(0,1));
    Trace(buffer);
    sprintf(buffer,"%f %f ",rr->A(1,1),rr->d(0,1));
    Trace(buffer);

    sprintf(buffer,"%f ",rr->centroid_x);
    Trace(buffer);
    sprintf(buffer,"%f ",rr->centroid_y);
    Trace(buffer);

    Trace("\n");
    i++;
  }
}

void ConnectedComponentsCorrespondence(
  const ImgInt& in, 
  const ImgInt& in_connect,
  std::vector<int>* connect_out,
  std::vector<int>* correspond_out)
{
  assert(in.Width() == in_connect.Width());
  assert(in.Height() == in_connect.Height());

  std::vector<int> found;
  std::vector<int>::const_iterator f_p;

  ImgInt::ConstIterator in_p, in_c_p;
  in_p = in.Begin();
  in_c_p = in_connect.Begin();
  int index = 0;
  while(in_p != in.End())
  {
    bool value_present = false;
    f_p = found.begin();
    while(f_p != found.end())
    {
      if(*f_p == *in_c_p)
      {
        value_present = true; 
        break;
      }
      f_p++;
    }
    if(value_present == false)
    {
      correspond_out->push_back(*in_p);
      connect_out->push_back(*in_c_p);
      found.push_back(*in_c_p);
      if(*in_c_p != index) 
        AfxMessageBox("Warning: ConnectedComponents labels not in order");
      index++;
    }
    in_p++;
    in_c_p++;
  }
}

void ReplacePixels(const ImgInt& in, int value_find, int value_replace, ImgInt* out)
{
  out->Reset(in.Width(),in.Height());
  *out = in;

  ImgInt::ConstIterator in_p;
  ImgInt::Iterator out_p = out->Begin();
  for(in_p=in.Begin(); in_p!=in.End(); in_p++)
  {
    if(*in_p == value_find)
      *out_p = value_replace;
    out_p++;
  }
}

void ReplacePixels(
  const ImgInt& in, 
  const std::vector<int>& find,
  const std::vector<int>& replace,
  ImgInt* out)
{

  assert(find.size() == replace.size());

  out->Reset(in.Width(),in.Height());
  *out = in;

  std::vector<int>::const_iterator f_p, r_p;
  ImgInt::Iterator out_p = out->Begin();
  ImgInt::ConstIterator in_p;
  for(in_p=in.Begin(); in_p!=in.End(); in_p++)
  {
    f_p = find.begin();
    r_p = replace.begin();
    while(f_p != find.end())
    {
      if(*in_p == *f_p)
      {
        *out_p = *r_p;
        break;
      }
      
      f_p++;
      r_p++;
    }

    out_p++;
  }
}

bool IsNeighbor(const ImgInt& in, int value_find, int value_neighbor)
{
  assert(in.Width() > 2);
  assert(in.Height() > 2);

  int wt = in.Width();
  int ht = in.Height();

  int i, j, k, i_s, i_e, j_s, j_e;
  ImgInt::ConstIterator in_p = in.Begin();
  for(j=0; j<ht; j++)
    for(i=0; i<wt; i++)
    {
      if(*in_p == value_find)
      {
        i_s = i;
        for(k=0; k<1; k++)
        {
          if(i_s == 0) break;
          i_s--;
        }
        i_e = i;
        for(k=0; k<=1; k++)
        {
          i_e++;
          if(i_e == wt) break;
        }
        j_s = j;
        for(k=0; k<1; k++)
        {
          if(j_s == 0) break;
          j_s--;
        }
        j_e = j;
        for(k=0; k<=1; k++)
        {
          j_e++;
          if(j_e == ht) break;
        }
      
        Rect box = Rect(i_s,j_s,i_e,j_e);
        ImgInt::ConstRectIterator box_p = in.BeginRect(box);
        while(!box_p.AtEnd())
        {
          if(*box_p == value_neighbor)
            return true;
          box_p++;
        }
      }

      in_p++;
    }

  return false;
}

bool IsNeighborCCImage(const ImgInt& in, int value_find, int value_neighbor)
{
  std::vector<Point> boundary;
  BoundaryOut(in, value_find, &boundary);

  std::vector<Point>::const_iterator by_p;
  for(by_p=boundary.begin(); by_p!=boundary.end(); by_p++)
    if(in(by_p->x,by_p->y) == value_neighbor)
      return true;

  return false;
}

void RemoveSmallRegions(const ImgInt& in, const int area_limit, ImgInt* out)
{

  ImgInt in_reorder;
  ConnectedComponents(in,&in_reorder);

  int min_value = Min(in_reorder);
  int max_value = Max(in_reorder);

  std::vector<int> region_pixel_values;
  std::vector<int> replacement_values;
  
  int reorder_value = 1;
  for(int i=min_value; i<=max_value; i++)
  {
    region_pixel_values.push_back(i);
    if(PixelFrequency(in_reorder,i) <= area_limit)
      replacement_values.push_back(0);
    else
    {
      replacement_values.push_back(reorder_value);
      reorder_value++;
    }
  }

  ReplacePixels(in_reorder,region_pixel_values,replacement_values,out);
}

void BoundaryIn(const ImgInt& in, const int find, std::vector<Point>* out)
{

  int wt = in.Width();
  int ht = in.Height();

  int max_val = Max(in);

  ImgInt tmp;
  tmp.Reset(wt+2,ht+2);
  Set(&tmp,(max_val+1));

  ImgInt::ConstIterator in_p = in.Begin();
  Rect box(1,1,wt+1,ht+1);
  ImgInt::RectIterator tmp_p = tmp.BeginRect(box);
  while(!tmp_p.AtEnd())
  {
    *tmp_p = *in_p;
    tmp_p++;
    in_p++;
  }

  int tmp_x, tmp_y;
  PickFirstPixel(tmp,find,&tmp_x,&tmp_y);
  int start_x = tmp_x;
  int start_y = tmp_y;

  int dir = 0;
  std::vector<Point> tmp_points;
  do
  {
    tmp_points.push_back(Point(tmp_x,tmp_y));
    if(dir == 0) // left
    {
      if(tmp(tmp_x,tmp_y-1) == find) { dir = 1; tmp_y--; }
      else if(tmp(tmp_x-1,tmp_y) != find) { dir = 3; tmp_y++; }
      else { dir = 0; tmp_x--; }
    }
    else if(dir == 1) // top
    {
      if(tmp(tmp_x+1,tmp_y) == find) { dir = 2; tmp_x++; }
      else if(tmp(tmp_x,tmp_y-1) != find) { dir = 0; tmp_x--; }
      else { dir = 1; tmp_y--; }
    }
    else if(dir == 2) // right
    {
      if(tmp(tmp_x,tmp_y+1) == find) { dir = 3; tmp_y++; }
      else if(tmp(tmp_x+1,tmp_y) != find) { dir = 1; tmp_y--; }
      else { dir = 2; tmp_x++; }
    }
    else // bottom
    {
      if(tmp(tmp_x-1,tmp_y) == find) { dir = 0; tmp_x--; }
      else if(tmp(tmp_x,tmp_y+1) != find) { dir = 2; tmp_x++; }
      else { dir = 3; tmp_y++; }
    }
  }while((tmp_x!=start_x) || (tmp_y!=start_y));

  int out_x, out_y;
  std::vector<Point>::const_iterator tmp_points_p;
  for(tmp_points_p=tmp_points.begin(); tmp_points_p!=tmp_points.end(); tmp_points_p++)
  {
    out_x = tmp_points_p->x;
    out_y = tmp_points_p->y;
    if((out_x>0) && (out_x<(wt+1)) && (out_y>0) && (out_y<(ht+1)))
      out->push_back(Point(out_x-1,out_y-1));
  }
}

// does not venture into single pixel thick crevices inside the region
void BoundaryOut(const ImgInt& in, const int find, std::vector<Point>* out)
{

  int wt = in.Width();
  int ht = in.Height();

  int max_val = Max(in);

  ImgInt tmp;
  tmp.Reset(wt+4,ht+4);
  Set(&tmp,(max_val+1));

  ImgInt::ConstIterator in_p = in.Begin();
  Rect box(2,2,wt+2,ht+2);
  ImgInt::RectIterator tmp_p = tmp.BeginRect(box);
  while(!tmp_p.AtEnd())
  {
    *tmp_p = *in_p;
    tmp_p++;
    in_p++;
  }

  struct DirectionPtr
  {
    int x;
    int y;
    DirectionPtr* previous;
    DirectionPtr* next;
  } *left, *bottom, *right, *top, *direction;

  left = new DirectionPtr; 
  left->x = -1;
  left->y = 0;
  bottom = new DirectionPtr;
  bottom->x = 0;
  bottom->y = 1;
  right = new DirectionPtr;
  right->x = 1;
  right->y = 0;
  top = new DirectionPtr;
  top->x = 0;
  top->y = -1;

  // counterclockwise connection
  left->next = bottom; left->previous = top;
  bottom->next = right; bottom->previous = left;
  right->next = top; right->previous = bottom;
  top->next = left; top->previous = right;

  int loc_x, loc_y;
  PickFirstPixel(tmp,find,&loc_x,&loc_y);
  loc_x--;
  int start_x = loc_x;
  int start_y = loc_y;

  direction = left;
  std::vector<Point> tmp_points;
  do
  {
    //TRACE("x,y is %d, %d\n",loc_x-2,loc_y-2);
    tmp_points.push_back(Point(loc_x,loc_y));
    if(tmp(loc_x + direction->x + direction->previous->x,
           loc_y + direction->y + direction->previous->y) == find)
      direction = direction->previous;
    else if((tmp(loc_x + direction->next->x, loc_y + direction->next->y) != find) &&
            (tmp(loc_x + direction->x + direction->next->x,
                 loc_y + direction->y + direction->next->y) != find) &&
             tmp(loc_x + direction->x, loc_y + direction->y) != find)
      direction = direction->next;
    else if(tmp(loc_x + direction->x, loc_y + direction->y) == find)
      direction = direction->previous;

    loc_x = loc_x + direction->x;
    loc_y = loc_y + direction->y;
  }while((loc_x!=start_x) || (loc_y!=start_y));

  delete left;
  delete bottom;
  delete right;
  delete top;

  int out_x, out_y;
  std::vector<Point>::const_iterator tmp_points_p;
  for(tmp_points_p=tmp_points.begin(); tmp_points_p!=tmp_points.end(); tmp_points_p++)
  {
    out_x = tmp_points_p->x;
    out_y = tmp_points_p->y;
    if((out_x>1) && (out_x<(wt+2)) && (out_y>1) && (out_y<(ht+2)))
      out->push_back(Point(out_x-2,out_y-2));
  }
}

void BoundaryOutArray(
  const ImgInt& in, 
  const std::vector<int>& find, 
  std::vector<PointArray>* out)
{

  int wt = in.Width();
  int ht = in.Height();

  int max_val = Max(in);

  ImgInt tmp;
  tmp.Reset(wt+4,ht+4);
  Set(&tmp,(max_val+1));

  ImgInt::ConstIterator in_p = in.Begin();
  Rect box(2,2,wt+2,ht+2);
  ImgInt::RectIterator tmp_p = tmp.BeginRect(box);
  while(!tmp_p.AtEnd())
  {
    *tmp_p = *in_p;
    tmp_p++;
    in_p++;
  }

  struct DirectionPtr
  {
    int x;
    int y;
    DirectionPtr* previous;
    DirectionPtr* next;
  } *left, *bottom, *right, *top, *direction;

  left = new DirectionPtr; 
  left->x = -1;
  left->y = 0;
  bottom = new DirectionPtr;
  bottom->x = 0;
  bottom->y = 1;
  right = new DirectionPtr;
  right->x = 1;
  right->y = 0;
  top = new DirectionPtr;
  top->x = 0;
  top->y = -1;

  // counterclockwise connection
  left->next = bottom; left->previous = top;
  bottom->next = right; bottom->previous = left;
  right->next = top; right->previous = bottom;
  top->next = left; top->previous = right;

  std::vector<Point> boundary_out;
  std::vector<int>::const_iterator find_p;
  for(find_p=find.begin(); find_p!=find.end(); find_p++)
  {
    int loc_x, loc_y;
    PickFirstPixel(tmp,(*find_p),&loc_x,&loc_y);
    loc_x--;
    int start_x = loc_x;
    int start_y = loc_y;

    direction = left;
    std::vector<Point> tmp_points;
    do
    {
      //TRACE("x,y is %d, %d\n",loc_x-2,loc_y-2);
      tmp_points.push_back(Point(loc_x,loc_y));
      if(tmp(loc_x + direction->x + direction->previous->x,
             loc_y + direction->y + direction->previous->y) == (*find_p))
        direction = direction->previous;
      else if((tmp(loc_x + direction->next->x, loc_y + direction->next->y) != (*find_p)) &&
              (tmp(loc_x + direction->x + direction->next->x,
                   loc_y + direction->y + direction->next->y) != (*find_p)) &&
               tmp(loc_x + direction->x, loc_y + direction->y) != (*find_p))
        direction = direction->next;
      else if(tmp(loc_x + direction->x, loc_y + direction->y) == (*find_p))
        direction = direction->previous;

      loc_x = loc_x + direction->x;
      loc_y = loc_y + direction->y;
    }while((loc_x!=start_x) || (loc_y!=start_y));

    int out_x, out_y;
    std::vector<Point>::const_iterator tmp_points_p;
    for(tmp_points_p=tmp_points.begin(); tmp_points_p!=tmp_points.end(); tmp_points_p++)
    {
      out_x = tmp_points_p->x;
      out_y = tmp_points_p->y;
      if((out_x>1) && (out_x<(wt+2)) && (out_y>1) && (out_y<(ht+2)))
        boundary_out.push_back(Point(out_x-2,out_y-2));
    }
    out->push_back(boundary_out);
  }

  delete left;
  delete bottom;
  delete right;
  delete top;
}


void FrequencyOrder(const std::vector<int>& in, std::vector<int>* out)
{
  int *in_raw;
  in_raw = new int[in.size()];
  std::vector<int>::const_iterator in_p;
  int itr = 0;
  for(in_p = in.begin(); in_p != in.end(); in_p++)
  {
    *(in_raw + itr) = *in_p;
    itr++;
  }

  qsort(in_raw, in.size(), sizeof(int), compare);

  std::vector<int> value_ordered;
  std::vector<int> value_ordered_frequency; 
  int previous_value = *in_raw;
  int frequency = 0;
  for(itr=0; itr<in.size(); itr++)
  {
    if(*(in_raw + itr) != previous_value)
    {
      value_ordered.push_back(previous_value);
      value_ordered_frequency.push_back(frequency);
      previous_value = *(in_raw + itr);
      frequency = 0;
    }
    frequency++;
  }
  if(*(in_raw + itr - 1) == previous_value)
  {
    value_ordered.push_back(previous_value);
    value_ordered_frequency.push_back(frequency);
  }

  delete[] in_raw;

  int *frequency_ordered_raw;
  int for_size = value_ordered_frequency.size();
  frequency_ordered_raw = new int[for_size];
  
  std::vector<int>::const_iterator vof_cp;
  itr = 0;
  for(vof_cp=value_ordered_frequency.begin(); vof_cp!=value_ordered_frequency.end(); vof_cp++)
  {
    *(frequency_ordered_raw + itr) = *vof_cp;
    itr++;
  }

  qsort(frequency_ordered_raw, for_size, sizeof(int), comparereverse);

  //TraceIntArray(value_ordered);
  //TraceIntArray(value_ordered_frequency);
  //TraceIntArray(frequency_ordered_raw,for_size);

  std::vector<int>::iterator vo_p, vof_p;
  for(itr=0; itr<for_size; itr++)
  {
    vo_p = value_ordered.begin();
    vof_p = value_ordered_frequency.begin();
    while(vof_p != value_ordered_frequency.end())
    {
      if((*(frequency_ordered_raw+itr)) == (*vof_p))
      {
        out->push_back(*vo_p);
        value_ordered.erase(vo_p);
        value_ordered_frequency.erase(vof_p);
        break;
      }
      vo_p++;
      vof_p++;
    }
  }

  delete[] frequency_ordered_raw;

}

// connected components input image expected
// output not connected components image
void ShrinkImage(const ImgInt& in, const int shrink_width, const int shrink_height, ImgInt* out)
{
  assert(shrink_width%2 == 0); // must be even
  assert(shrink_height%2 == 0); // must be even
  assert(in.Width() > shrink_width);
  assert(in.Height() > shrink_height);

  out->Reset(in.Width(), in.Height());
  Set(out, 0);
  Rect box((shrink_width/2), (shrink_height/2),
           (in.Width() - (shrink_width/2)), (in.Height() - (shrink_height/2)));

  ImgInt::ConstRectIterator in_p = in.BeginRect(box);
  ImgInt::RectIterator out_p=out->BeginRect(box);
  while(!in_p.AtEnd())
  {
    *out_p = *in_p + 1;
    out_p++;
    in_p++;
  }
}

// does not venture into single pixel thick crevices inside the region
void RegionBoundariesOut(
  const ImgInt& in, 
  const std::vector<int>& find,
  const std::vector<Point>& find_first_point,
  std::vector<PointVect>* boundaries,
  std::vector<IntVect>* boundary_values,
  std::vector<bool>* unique_boundary)
{
  assert(find.size() == find_first_point.size());

  int wt = in.Width();
  int ht = in.Height();

  int min_val = Min(in);
  int max_val = Max(in);

  assert(min_val == 0);
  assert(max_val <= max_connectedcomponents_label);

  ImgInt tmp;
  tmp.Reset(wt+4,ht+4);
  Set(&tmp,(max_val+1));

  ImgInt::ConstIterator in_p = in.Begin();
  Rect box(2,2,wt+2,ht+2);
  ImgInt::RectIterator tmp_p = tmp.BeginRect(box);
  while(!tmp_p.AtEnd())
  {
    *tmp_p = *in_p;
    tmp_p++;
    in_p++;
  }

  struct DirectionPtr
  {
    int x;
    int y;
    DirectionPtr* previous;
    DirectionPtr* next;
  } *left, *bottom, *right, *top, *direction;

  left = new DirectionPtr; 
  left->x = -1;
  left->y = 0;
  bottom = new DirectionPtr;
  bottom->x = 0;
  bottom->y = 1;
  right = new DirectionPtr;
  right->x = 1;
  right->y = 0;
  top = new DirectionPtr;
  top->x = 0;
  top->y = -1;

  // counterclockwise connection
  left->next = bottom; left->previous = top;
  bottom->next = right; bottom->previous = left;
  right->next = top; right->previous = bottom;
  top->next = left; top->previous = right;

  std::vector<int>::const_iterator f_p;
  std::vector<Point>::const_iterator f_fp_p;
  f_p = find.begin();
  f_fp_p = find_first_point.begin();
  while(f_p != find.end())
  {
    int loc_x = (f_fp_p->x) + 2;
    int loc_y = (f_fp_p->y) + 2;
    
    loc_x--;
    int start_x = loc_x;
    int start_y = loc_y;

    direction = left;
    std::vector<Point> tmp_points;
    do
    {
      tmp_points.push_back(Point(loc_x,loc_y));
      if(tmp(loc_x + direction->x + direction->previous->x,
             loc_y + direction->y + direction->previous->y) == (*f_p))
        direction = direction->previous;
      else if((tmp(loc_x + direction->next->x, loc_y + direction->next->y) != (*f_p)) &&
              (tmp(loc_x + direction->x + direction->next->x,
                   loc_y + direction->y + direction->next->y) != (*f_p)) &&
               tmp(loc_x + direction->x, loc_y + direction->y) != (*f_p))
        direction = direction->next;
      else if(tmp(loc_x + direction->x, loc_y + direction->y) == (*f_p))
        direction = direction->previous;

      loc_x = loc_x + direction->x;
      loc_y = loc_y + direction->y;
    }while((loc_x!=start_x) || (loc_y!=start_y));

    int out_x, out_y;
    std::vector<Point> out;
    std::vector<int> out_values;
    bool unique = true;
    int val, val_old = -1;
    std::vector<Point>::const_iterator tp_p;
    for(tp_p=tmp_points.begin(); tp_p!=tmp_points.end(); tp_p++)
    {
      out_x = tp_p->x;
      out_y = tp_p->y;
      if((out_x>1) && (out_x<(wt+2)) && (out_y>1) && (out_y<(ht+2)))
      {
        out.push_back(Point(out_x-2,out_y-2));
        val = tmp(out_x,out_y);
        out_values.push_back(val);
        if(val_old != -1 && val_old != val)
          unique = false;
        val_old = val;
      }
    }

    boundaries->push_back(out);
    boundary_values->push_back(out_values);
    unique_boundary->push_back(unique);

    f_p++;
    f_fp_p++;
  }

  delete left;
  delete bottom;
  delete right;
  delete top;
}

void ConnectedComponentsMyInfoOut(
  const ImgInt& in, 
  const ImgInt& in_connect,
  std::vector<int>* connect_out,
  std::vector<int>* connect_out_npixels,
  std::vector<Point>* connect_out_first_point,
  std::vector<int>* correspond_out)
{
  assert(in.Width() == in_connect.Width());
  assert(in.Height() == in_connect.Height());

  int min_label = Min(in_connect);
  int max_label = Max(in_connect);

  assert(min_label == 0);
  assert(max_label <= max_connectedcomponents_label);

  connect_out->resize(max_label+1);
  connect_out_npixels->resize(max_label+1);
  connect_out_first_point->resize(max_label+1);
  correspond_out->resize(max_label+1);

  std::vector<bool> found;
  found.resize(max_label+1);

  // initializing all the arrays  
  int index = 0;
  std::vector<int>::iterator conn_p, conn_np_p, corr_p;
  std::vector<Point>::iterator conn_fp_p;
  std::vector<bool>::iterator f_p;
  conn_p = connect_out->begin();
  conn_np_p = connect_out_npixels->begin();
  conn_fp_p = connect_out_first_point->begin();
  corr_p = correspond_out->begin();
  f_p = found.begin();
  while(conn_p != connect_out->end())
  {
    *conn_p = index;
    *conn_np_p = 0;
    *conn_fp_p = Point(0,0);
    *corr_p = 0;
    *f_p = false;

    conn_p++;
    conn_np_p++;
    conn_fp_p++;
    corr_p++;
    f_p++;
    index++;
  }

  ImgInt::ConstIterator in_p, in_c_p;
  in_p = in.Begin();
  in_c_p = in_connect.Begin();
  int x = 0, y = 0;
  while(in_p != in.End())
  {
    if((*in_c_p)>=0 && (*in_c_p)<=max_label)
    {
      (*connect_out_npixels)[*in_c_p]++;
      if(found[*in_c_p] == false)
      {
        (*connect_out_first_point)[*in_c_p] = Point(x,y);
        (*correspond_out)[*in_c_p] = (*in_p);
        found[*in_c_p] = true;
      }
      in_p++;
      in_c_p++;
      x++;
      if(x == in.Width())
      {
        x = 0;
        y++;
      }
    }
    else
    {
      AfxMessageBox("Error: invalid label in ConnectedComponentsMy image");
      assert(0);
    }
  }
}

void DiffuseRegionsDebug(const ImgInt& in, const int area_limit, ImgInt* out, std::vector<int>* previous_labels)
{

  ImgInt in_reorder;
  ConnectedComponents(in,&in_reorder);

  std::vector<int> reorder_region_pixel_values;
  std::vector<int> in_pixel_values;
  ConnectedComponentsCorrespondence(in,in_reorder,&reorder_region_pixel_values,&in_pixel_values);

  int reorder_min_value = Min(in_reorder);
  int reorder_max_value = Max(in_reorder);

  std::vector<int> reorder_replacement_values;
  int reorder_value = 1;
  for(int i=reorder_min_value; i<=reorder_max_value; i++)
  {
    if(PixelFrequency(in_reorder,i) <= area_limit)
      reorder_replacement_values.push_back(0);
    else
    {
      reorder_replacement_values.push_back(reorder_value);
      reorder_value++;
    }
  }

  ImgInt shed_unconnect;
  ReplacePixels(in_reorder,reorder_region_pixel_values,reorder_replacement_values,&shed_unconnect);
  
  ImgInt shed;
  ConnectedComponents(shed_unconnect, &shed);

  std::vector<int> shed_pixel_values;
  reorder_region_pixel_values.resize(0);
  ConnectedComponentsCorrespondence(shed,in_reorder,&reorder_region_pixel_values,&shed_pixel_values);

  //TraceIntArray(in_pixel_values);
  //TraceIntArray(reorder_region_pixel_values);
  //TraceIntArray(shed_pixel_values);

  //Figure fig1("in_reorder");
  //fig1.Draw(in_reorder);
  //Figure fig2("shed");
  //fig2.Draw(shed);
  //AfxMessageBox("press ok to continue");

  std::vector<int> region_pixel_values;
  std::vector<int> replacement_values;
  std::vector<Point> boundary_points;
  std::vector<int> boundary_values;
  std::vector<int> neighbor_values;

  std::vector<int>::iterator sd_p, iv_p;  
  std::vector<int>::const_iterator nr_p;
  int min_value = Min(shed);
  int max_value = Max(shed);
  for(i=min_value; i<=max_value; i++)
  {
    if(PixelFrequency(shed,i) <= area_limit)
    {
      int itr;
      do
      {
        iv_p = in_pixel_values.begin();
        sd_p = shed_pixel_values.begin();
        itr = 0;
        while(sd_p != shed_pixel_values.end())
        {
          if(i == *sd_p)
          {
            in_pixel_values.erase(iv_p);
            shed_pixel_values.erase(sd_p);
            break;
          }
          iv_p++;
          sd_p++;
          itr++;
          //TRACE("\n val is %d and itr is %d\n",i,itr);
        }
      }while(itr != shed_pixel_values.size());

      region_pixel_values.push_back(i);
      //TRACE(" replacing label %d with",i);
      boundary_points.resize(0);
      boundary_values.resize(0);
      neighbor_values.resize(0);
      BoundaryOut(shed,i,&boundary_points);
      GetPixelValues(shed,boundary_points,&boundary_values);
      FrequencyOrder(boundary_values,&neighbor_values);
      for(nr_p=neighbor_values.begin(); nr_p!=neighbor_values.end(); nr_p++)
      {
        if(PixelFrequency(shed,(*nr_p)) > area_limit)
        {
          replacement_values.push_back(*nr_p);
          //TRACE(" label %d\n",*nr_p);
          break;
        }
      }
    }
  }

  //TraceIntArray(region_pixel_values);
  //TraceIntArray(replacement_values);

  //TraceIntArray(in_pixel_values);
  //TraceIntArray(shed_pixel_values);

  ImgInt out_unorder;
  ReplacePixels(shed,region_pixel_values,replacement_values,&out_unorder);
  ConnectedComponents(out_unorder,out);

  std::vector<int> out_labels_correspond;
  std::vector<int> out_labels;
  ConnectedComponentsCorrespondence(out_unorder,*out,&out_labels,&out_labels_correspond);

  //TraceIntArray(out_labels);
  //TraceIntArray(out_labels_correspond);

  std::vector<int>::const_iterator olc_p;
  for(olc_p=out_labels_correspond.begin(); olc_p!=out_labels_correspond.end(); olc_p++)
  {
    iv_p = in_pixel_values.begin();
    sd_p = shed_pixel_values.begin();
    while(sd_p != shed_pixel_values.end())
    {
      if(*olc_p == *sd_p)
      {
        previous_labels->push_back(*iv_p);
        break;
      }
      iv_p++;
      sd_p++;
    }
  }

  //TraceIntArray(*previous_labels);
   
}

void DiffuseRegions(const ImgInt& in, const int area_limit, ImgInt* out)
{

  ImgInt in_reorder;
  ConnectedComponents(in,&in_reorder);

  int reorder_min_value = Min(in_reorder);
  int reorder_max_value = Max(in_reorder);

  std::vector<int> reorder_region_pixel_values;
  std::vector<int> reorder_replacement_values;
  
  int reorder_value = 1;
  for(int i=reorder_min_value; i<=reorder_max_value; i++)
  {
    reorder_region_pixel_values.push_back(i);
    if(PixelFrequency(in_reorder,i) <= area_limit)
      reorder_replacement_values.push_back(0);
    else
    {
      reorder_replacement_values.push_back(reorder_value);
      reorder_value++;
    }
  }

  ImgInt shed_unconnect;
  ReplacePixels(in_reorder,reorder_region_pixel_values,reorder_replacement_values,&shed_unconnect);
  
  ImgInt shed;
  ConnectedComponents(shed_unconnect, &shed);
  
  std::vector<int> region_pixel_values;
  std::vector<int> replacement_values;
  std::vector<Point> boundary_points;
  std::vector<int> boundary_values;
  std::vector<int> neighbor_values;
  
  std::vector<int>::const_iterator nr_p;
  int min_value = Min(shed);
  int max_value = Max(shed);
  for(i=min_value; i<=max_value; i++)
  {
    if(PixelFrequency(shed,i) <= area_limit)
    {
      region_pixel_values.push_back(i);
      boundary_points.resize(0);
      boundary_values.resize(0);
      neighbor_values.resize(0);
      BoundaryOut(shed,i,&boundary_points);
      GetPixelValues(shed,boundary_points,&boundary_values);
      FrequencyOrder(boundary_values,&neighbor_values);
      for(nr_p=neighbor_values.begin(); nr_p!=neighbor_values.end(); nr_p++)
      {
        if(PixelFrequency(shed,(*nr_p)) > area_limit)
        {
          replacement_values.push_back(*nr_p);
          break;
        }
      }
    }
  }

  ImgInt out_unorder;
  ReplacePixels(shed,region_pixel_values,replacement_values,&out_unorder);
  ConnectedComponents(out_unorder,out);

}

void ScaleFloatImage(
  const ImgFloat& in,
  ImgGray* out)
{
  float min_val = Min(in);
  float max_val = Max(in);

  ImgFloat tmp;
  tmp.Reset(in.Width(), in.Height());
  Set(out, 0.0f);

  float diff = max_val-min_val;
  if(diff > 0)
  {
    ImgFloat::ConstIterator in_p;
    ImgFloat::Iterator tmp_p;
    for(in_p=in.Begin(), tmp_p=tmp.Begin(); in_p!=in.End(); in_p++, tmp_p++)
      *tmp_p = ((*in_p)-min_val)*(255.0)/diff;
  }

  Convert(tmp, out);
}

void ScaleDisparities(
  const ImgFloat& in,
  const float scale_factor,
  ImgGray* out)
{
  ImgFloat tmp;
  tmp.Reset(in.Width(), in.Height());
  Set(out, 0.0f);

  ImgFloat::ConstIterator in_p;
  ImgFloat::Iterator tmp_p;
  for(in_p=in.Begin(), tmp_p=tmp.Begin(); in_p!=in.End(); in_p++, tmp_p++)
    *tmp_p = scale_factor*(*in_p);

  Convert(tmp, out);
}

void WriteCorrelationVector(
  const int* in,
  const length,
  const char* filename)
{
  std::ofstream result_file(filename);
  char buffer[100];

  for(int i=0; i<length; i++)
  {
    sprintf(buffer, "%d\n", in[i]);
    result_file << buffer;
  }

  result_file.close();
}

inline bool Equal(const MatFlt& cmp1, const MatFlt& cmp2)
{
  if(cmp1.Width() != cmp2.Width())
    return false;
  if(cmp1.Height() == cmp2.Height())
    return false;

  MatFlt::ConstIterator cmp1_p = cmp1.Begin();
  MatFlt::ConstIterator cmp2_p = cmp2.Begin();
  while(cmp1_p!=cmp1.End())
  {
    if(*cmp1_p != *cmp2_p)
      return false;

    cmp1_p++;
    cmp2_p++;
  }

  return true;
}

void AdjustEquivalence(std::vector<MatFlt>& in)
{
  for(int i=0; i<(in.size()-1); i++)
    for(int j=(i+1); j<(in.size()); j++)
    {
      if(Equal(in[i],in[j]))
      {
        for(int k=j+1; k<(in.size()); k++)
          in[k-1] = in[k];
      
        in.resize(in.size()-1);
      }
    }
}

void AdjustEquivalence(std::vector<float>& in)
{
  for(int i=0; i<(in.size()-1); i++)
    for(int j=(i+1); j<(in.size()); j++)
    {
      if(in[i] == in[j])
      {
        for(int k=j+1; k<(in.size()); k++)
          in[k-1] = in[k];
      
        in.resize(in.size()-1);
      }
    }
}
