#include <fstream>
#include "lucas.h"

typedef std::vector<int> IntVect;
typedef std::vector<float> FloatVect;
typedef std::vector<bool> BoolVect;
typedef std::vector<Point> PointVect;

void SeeInt(int val);

void SeeFloat(float val);

void SeeLocation(Location p);

void WriteFloatVect(
  char* filename,
  const std::vector<float>& flt_vect);

float MeanFloatVect(const std::vector<float>& flt_vect);

void MedianFilterFinal(
  const ImgInt& in, 
  const int filter_width, const int filter_height, 
  ImgInt* out);

void WriteResults(
  const std::vector<MatFlt>& mat,
  const std::vector<float>& centroid_x,
  const std::vector<float>& centroid_y,
  const std::vector<int>& status,
  const char* filename);

void TraceDispFunc(
  const std::vector<MatFlt>& mat,
  const std::vector<float>& centroid_x,
  const std::vector<float>& centroid_y);

void WriteResults(
  const std::vector<MatFlt>& mat,
  const std::vector<float>& centroid_x,
  const std::vector<float>& centroid_y,
  const std::vector<float>& residue,
  const std::vector<int>& status,
  const char* filename);

void WriteResults(
  const std::vector<LKAResult>& result_r,
  const char* filename);

void TraceLKAResultArray(const std::vector<LKAResult>& result_r);

// inline small functions
inline void AddValue(const ImgInt& in, const int value, ImgInt* out)
{
  out->Reset(in.Width(),in.Height());

  ImgInt::Iterator out_p = out->Begin();
  ImgInt::ConstIterator in_p;
  for(in_p=in.Begin(); in_p!=in.End(); in_p++)
  {
    *out_p = (*in_p) + value;
    out_p++;
  }
}

inline void SubtractValue(const ImgInt& in, const int value, ImgInt* out)
{
  out->Reset(in.Width(),in.Height());

  ImgInt::Iterator out_p = out->Begin();
  ImgInt::ConstIterator in_p;
  for(in_p=in.Begin(); in_p!=in.End(); in_p++)
  {
    *out_p = (*in_p) - value;
    out_p++;
  }
}

inline int PixelFrequency(const ImgInt& in, const int value)
{
  int frequency = 0;
  ImgInt::ConstIterator in_p;
  for(in_p=in.Begin(); in_p!=in.End(); in_p++)
    if(*in_p == value)
      frequency++;

  return frequency;
}

inline void GetPixelValues(const ImgInt& in, const std::vector<Point>& loc, std::vector<int>* out)
{
  std::vector<Point>::const_iterator loc_p;
  for(loc_p=loc.begin(); loc_p!=loc.end(); loc_p++)
    out->push_back(*in.Begin(loc_p->x, loc_p->y));
}

inline void SetPixels(const ImgInt& in, const std::vector<Point>& loc, const int replace, ImgInt* out)
{
  out->Reset(in.Width(),in.Height());
  *out = in;

  std::vector<Point>::const_iterator loc_p;
  for(loc_p=loc.begin(); loc_p!=loc.end(); loc_p++)
    *(out->Begin(loc_p->x,loc_p->y)) = replace;
}

inline void PickFirstPixel(const ImgInt& in, int value, int* out_x, int* out_y)
{
  ImgInt::ConstIterator p;
  int i=0,j=0; 
  for(p=in.Begin(); p!=in.End(); p++)
  {
    if(*p == value)
    {
      *out_x = i;
      *out_y = j;
      break;
    }

    i++;
    if(i == in.Width())
    {
      i=0;
      j++;
    }
  }

  if(p==in.End()) AfxMessageBox("Warning: PickFirstPixel failed\n");
}

inline bool SetPixelsBinary(const ImgInt& in, const int find1, const int find2, ImgBinary* out)
{
  out->Reset(in.Width(),in.Height());
  Set(out,false);

  bool flag = false;

  ImgInt::ConstIterator in_p;
  ImgBinary::Iterator out_p = out->Begin();
  for(in_p=in.Begin(); in_p!=in.End(); in_p++)
  {
    if((*in_p==find1) || (*in_p==find2))
    {
      *out_p = true;
      flag = true;
    }
    out_p++;
  }
  return flag;
}

// regular functions

void ConnectedComponentsCorrespondence(
  const ImgInt& in, 
  const ImgInt& in_connect,
  std::vector<int>* connect_out,
  std::vector<int>* correspond_out);

void ReplacePixels(const ImgInt& in, int value_find, int value_replace, ImgInt* out);

void ReplacePixels(
  const ImgInt& in, 
  const std::vector<int>& find, 
  const std::vector<int>& replace,
  ImgInt* out);

bool IsNeighbor(const ImgInt& in, int value_find, int value_neighbor);
bool IsNeighborCCImage(const ImgInt& in, int value_find, int value_neighbor);
void RemoveSmallRegions(const ImgInt& in, const int size_limit, ImgInt* out);
void BoundaryIn(const ImgInt& in, const int find, std::vector<Point>* out);
void BoundaryOut(const ImgInt& in, const int find, std::vector<Point>* out);

typedef std::vector<Point> PointArray;

void BoundaryOutArray(
  const ImgInt& in, 
  const std::vector<int>& find, 
  std::vector<PointArray>* out);

void FrequencyOrder(const std::vector<int>& in, std::vector<int>* out);
void ShrinkImage(const ImgInt& in, const int shrink_width, const int shrink_height, ImgInt* out);
void DiffuseRegionsDebug(const ImgInt& in, const int area_limit, ImgInt* out, std::vector<int>* previous_labels);

void RegionBoundariesOut(
  const ImgInt& in, 
  const std::vector<int>& find,
  const std::vector<Point>& find_first_point,
  std::vector<PointVect>* boundaries,
  std::vector<IntVect>* boundary_values,
  std::vector<bool>* unique_boundary);

void ConnectedComponentsMyInfoOut(
  const ImgInt& in, 
  const ImgInt& in_connect,
  std::vector<int>* connect_out,
  std::vector<int>* connect_out_npixels,
  std::vector<Point>* connect_out_first_point,
  std::vector<int>* correspond_out);

void ScaleFloatImage(
  const ImgFloat& in,
  ImgGray* out);

void ScaleDisparities(
  const ImgFloat& in,
  const float scale_factor,
  ImgGray* out);

void WriteCorrelationVector(
  const int* in,
  const length,
  const char* filename);

// other

bool Equal(const MatFlt& cmp1, const MatFlt& cmp2);
void AdjustEquivalence(std::vector<MatFlt>& in);
void AdjustEquivalence(std::vector<float>& in);

// debug functions
inline void TraceIntArray(const std::vector<int>& in)
{
  Trace("\n SIZE IS %d \n",in.size());

  std::vector<int>::const_iterator in_p;
  int index = 0;
  for(in_p=in.begin(); in_p!=in.end(); in_p++)
  {
    Trace(" index %d value %d \n",index,*in_p);
    index++;
  }
}

inline void TraceIntArray(const int* in, const int size)
{
  Trace("\n SIZE IS %d \n",size);

  int index = 0;
  while(index < size)
  {
    Trace(" index %d value %d \n",index,*(in+index));
    index++;
  }
}

inline void TracePointArray(const std::vector<Point>& in)
{
  Trace("\n SIZE IS %d \n",in.size());

  std::vector<Point>::const_iterator in_p;
  int index = 0;
  for(in_p=in.begin(); in_p!=in.end(); in_p++)
  {
    Trace(" index %d point (%d,%d) \n",index,in_p->x,in_p->y);
    index++;
  }
}

inline void WriteIntArray(const std::vector<int>& in, const char* file_name)
{
  std::ofstream out_file(file_name);

  char buffer[50];
  sprintf(buffer,"\n SIZE IS %d \n",in.size());
  out_file << buffer;

  std::vector<int>::const_iterator in_p;
  int index = 0;
  for(in_p=in.begin(); in_p!=in.end(); in_p++)
  {
    sprintf(buffer," index %d value %d \n",index,*in_p);
    out_file << buffer;
    index++;
  }

  out_file.close();
}

inline void TraceLKAParameters(const LKAParameters& params)
{
  Trace("\n");
  Trace("min_num_pixels = %d\n",params.min_num_pixels);
  Trace("num_iterations = %d\n",params.num_iterations);
  Trace("\noob_threshold = %f\n",params.oob_threshold);
  Trace("max_disp_threshold = %f\n",params.max_disp_threshold);
  Trace("min_disp_threshold = %f\n",params.min_disp_threshold);
  Trace("max_aff_threshold = %f\n",params.max_aff_threshold);
  Trace("min_aff_threshold = %f\n",params.min_aff_threshold);
  Trace("\ndet_threshold = %f\n",params.det_threshold);
  Trace("residue_threshold = %f\n",params.residue_threshold);
  Trace("\nshrink_width_ratio = %f\n",params.shrink_width_ratio);
  Trace("shrink_width_ratio = %f\n",params.shrink_height_ratio);
  Trace("smooth_sigma = %f\n",params.smooth_sigma);
  Trace("gradient_sigma = %f\n",params.gradient_sigma);
  Trace("gradient_kernel_length = %d\n",params.gradient_kernel_length);
  Trace("skip_label_zero = %d\n",params.skip_label_zero);
}
