#include "stdafx.h"
#include "../../../code/vc6/alpha1/src/blepo.h"
using namespace blepo;
#include "diffuse.h"

namespace
{

const int max_connectedcomponents_label = 10000;

typedef std::vector<int> intArray;
typedef std::vector<Point> PointArray;

int compare(const void * a, const void * b)
{
  if(*(int*)a < *(int*)b)
    return -1;
  else if(*(int*)a == *(int*)b)
    return 0;
  else
    return 1;
}

int MaxFrequencyIndex(const std::vector<int>& in)
{
  assert(in.size() > 0);

  int *in_raw;
  in_raw = new int[in.size()];
  std::vector<int>::const_iterator in_p;
  int itr = 0;
  for(in_p = in.begin(); in_p != in.end(); in_p++)
  {
    *(in_raw + itr) = *in_p;
    itr++;
  }

  qsort(in_raw, in.size(), sizeof(int), compare);

  std::vector<int> value_ordered;
  std::vector<int> value_ordered_frequency; 
  int previous_value = *in_raw;
  int frequency = 0;
  int max_frequency = 0;
  int max_frequency_value;
  for(itr=0; itr<in.size(); itr++)
  {
    if(*(in_raw + itr) != previous_value)
    {
      value_ordered.push_back(previous_value);
      value_ordered_frequency.push_back(frequency);
      if(frequency > max_frequency)
      {
        max_frequency = frequency;
        max_frequency_value = previous_value;
      }
      previous_value = *(in_raw + itr);
      frequency = 0;
    }
    frequency++;
  }
  if(*(in_raw + itr - 1) == previous_value)
  {
    value_ordered.push_back(previous_value);
    value_ordered_frequency.push_back(frequency);
    if(frequency > max_frequency)
    {
      max_frequency = frequency;
      max_frequency_value = previous_value;
    }
  }

  delete[] in_raw;

  return max_frequency_value;
}

void ValueToReplace(
  const ImgInt& in,
  const std::vector<int>& find,
  const std::vector<intArray>& boundary_values,
  const std::vector<bool>& unique_boundary,
  std::vector<int>* replace)
{
  assert(find.size() == boundary_values.size());
  assert(find.size() == unique_boundary.size());

  int min_val = Min(in);
  int max_val = Max(in);

  assert(min_val == 0);
  assert(max_val <= max_connectedcomponents_label);

  std::vector<int>::const_iterator f_p = find.begin();
  std::vector<intArray>::const_iterator bvs_p = boundary_values.begin();
  std::vector<bool>::const_iterator ub_p = unique_boundary.begin();
  while(f_p != find.end())
  {
    if(*ub_p == true)
      replace->push_back((*bvs_p)[0]);
    else
      replace->push_back(MaxFrequencyIndex(*bvs_p));     

    f_p++;
    bvs_p++;
    ub_p++;
  }
}

// does not venture into single pixel thick crevices inside the region
void RegionBoundaries(
  const ImgInt& in, 
  const std::vector<int>& find,
  const std::vector<Point>& find_first_point,
  std::vector<PointArray>* boundaries,
  std::vector<intArray>* boundary_values,
  std::vector<bool>* unique_boundary)
{
  assert(find.size() == find_first_point.size());

  int wt = in.Width();
  int ht = in.Height();

  int min_val = Min(in);
  int max_val = Max(in);

  assert(min_val == 0);
  assert(max_val <= max_connectedcomponents_label);

  ImgInt tmp;
  tmp.Reset(wt+4,ht+4);
  Set(&tmp,(max_val+1));

  ImgInt::ConstIterator in_p = in.Begin();
  Rect box(2,2,wt+2,ht+2);
  ImgInt::RectIterator tmp_p = tmp.BeginRect(box);
  while(!tmp_p.AtEnd())
  {
    *tmp_p = *in_p;
    tmp_p++;
    in_p++;
  }

  struct DirectionPtr
  {
    int x;
    int y;
    DirectionPtr* previous;
    DirectionPtr* next;
  } *left, *bottom, *right, *top, *direction;

  left = new DirectionPtr; 
  left->x = -1;
  left->y = 0;
  bottom = new DirectionPtr;
  bottom->x = 0;
  bottom->y = 1;
  right = new DirectionPtr;
  right->x = 1;
  right->y = 0;
  top = new DirectionPtr;
  top->x = 0;
  top->y = -1;

  // counterclockwise connection
  left->next = bottom; left->previous = top;
  bottom->next = right; bottom->previous = left;
  right->next = top; right->previous = bottom;
  top->next = left; top->previous = right;

  std::vector<int>::const_iterator f_p;
  std::vector<Point>::const_iterator f_fp_p;
  f_p = find.begin();
  f_fp_p = find_first_point.begin();
  while(f_p != find.end())
  {
    int loc_x = (f_fp_p->x) + 2;
    int loc_y = (f_fp_p->y) + 2;
    
    loc_x--;
    int start_x = loc_x;
    int start_y = loc_y;

    direction = left;
    std::vector<Point> tmp_points;
    do
    {
      tmp_points.push_back(Point(loc_x,loc_y));
      if(tmp(loc_x + direction->x + direction->previous->x,
             loc_y + direction->y + direction->previous->y) == (*f_p))
        direction = direction->previous;
      else if((tmp(loc_x + direction->next->x, loc_y + direction->next->y) != (*f_p)) &&
              (tmp(loc_x + direction->x + direction->next->x,
                   loc_y + direction->y + direction->next->y) != (*f_p)) &&
               tmp(loc_x + direction->x, loc_y + direction->y) != (*f_p))
        direction = direction->next;
      else if(tmp(loc_x + direction->x, loc_y + direction->y) == (*f_p))
        direction = direction->previous;

      loc_x = loc_x + direction->x;
      loc_y = loc_y + direction->y;
    }while((loc_x!=start_x) || (loc_y!=start_y));

    int out_x, out_y;
    std::vector<Point> out;
    std::vector<int> out_values;
    bool unique = true;
    int val, val_old = -1;
    std::vector<Point>::const_iterator tp_p;
    for(tp_p=tmp_points.begin(); tp_p!=tmp_points.end(); tp_p++)
    {
      out_x = tp_p->x;
      out_y = tp_p->y;
      if((out_x>1) && (out_x<(wt+2)) && (out_y>1) && (out_y<(ht+2)))
      {
        out.push_back(Point(out_x-2,out_y-2));
        val = tmp(out_x,out_y);
        out_values.push_back(val);
        if(val_old != -1 && val_old != val)
          unique = false;
        val_old = val;
      }
    }

    boundaries->push_back(out);
    boundary_values->push_back(out_values);
    unique_boundary->push_back(unique);

    f_p++;
    f_fp_p++;
  }

  delete left;
  delete bottom;
  delete right;
  delete top;
}

void ConnectedComponentsInfo(
  const ImgInt& in, 
  const ImgInt& in_connect,
  std::vector<int>* connect_out,
  std::vector<int>* connect_out_npixels,
  std::vector<Point>* connect_out_first_point,
  std::vector<int>* correspond_out)
{
  assert(in.Width() == in_connect.Width());
  assert(in.Height() == in_connect.Height());

  int min_label = Min(in_connect);
  int max_label = Max(in_connect);

  assert(min_label == 0);
  assert(max_label <= max_connectedcomponents_label);

  connect_out->resize(max_label+1);
  connect_out_npixels->resize(max_label+1);
  connect_out_first_point->resize(max_label+1);
  correspond_out->resize(max_label+1);

  std::vector<bool> found;
  found.resize(max_label+1);

  // initializing all the arrays  
  int index = 0;
  std::vector<int>::iterator conn_p, conn_np_p, corr_p;
  std::vector<Point>::iterator conn_fp_p;
  std::vector<bool>::iterator f_p;
  conn_p = connect_out->begin();
  conn_np_p = connect_out_npixels->begin();
  conn_fp_p = connect_out_first_point->begin();
  corr_p = correspond_out->begin();
  f_p = found.begin();
  while(conn_p != connect_out->end())
  {
    *conn_p = index;
    *conn_np_p = 0;
    *conn_fp_p = Point(0,0);
    *corr_p = 0;
    *f_p = false;

    conn_p++;
    conn_np_p++;
    conn_fp_p++;
    corr_p++;
    f_p++;
    index++;
  }

  ImgInt::ConstIterator in_p, in_c_p;
  in_p = in.Begin();
  in_c_p = in_connect.Begin();
  int x = 0, y = 0;
  while(in_p != in.End())
  {
    if((*in_c_p)>=0 && (*in_c_p)<=max_label)
    {
      (*connect_out_npixels)[*in_c_p]++;
      if(found[*in_c_p] == false)
      {
        (*connect_out_first_point)[*in_c_p] = Point(x,y);
        (*correspond_out)[*in_c_p] = (*in_p);
        found[*in_c_p] = true;
      }
      in_p++;
      in_c_p++;
      x++;
      if(x == in.Width())
      {
        x = 0;
        y++;
      }
    }
    else
    {
      AfxMessageBox("Error: invalid label in ConnectedComponents image");
      assert(0);
    }
  }
}

void ExtractRow(
  const ImgInt& in,
  const int j,
  std::vector<int>* row_val)
{
  row_val->resize(0);
  for(int i=0; i<in.Width(); i++)
    row_val->push_back(in(i,j));
}

void ExtractCol(
  const ImgInt& in,
  const int i,
  std::vector<int>* col_val)
{
  col_val->resize(0);
  for(int j=0; j<in.Height(); j++)
    col_val->push_back(in(i,j));
}

void ExtractRowZero(
  const std::vector<int>& row,
  std::vector<int>* row_zero_pos,
  std::vector<int>* row_zero_neg)
{
  row_zero_pos->resize(0);
  row_zero_neg->resize(0);
  std::vector<int>::const_iterator r_p;
  for(r_p=row.begin(); r_p!=row.end(); r_p++)
  {
    if(*r_p == 0)
    {
      row_zero_pos->push_back(1);
      row_zero_neg->push_back(0);
    }
    else
    {
      row_zero_pos->push_back(0);
      row_zero_neg->push_back(1);
    }
  }
}

void SetRow(
  const std::vector<int>& row,
  const int j,
  ImgInt& out)
{
  std::vector<int>::const_iterator r_p = row.begin();
  for(int i=0; i<out.Width(); i++)
  {
    out(i,j) = *r_p;
    r_p++;
  }
}

void SetCol(
  const std::vector<int>& col,
  const int i,
  ImgInt& out)
{
  std::vector<int>::const_iterator c_p = col.begin();
  for(int j=0; j<out.Height(); j++)
  {
    out(i,j) = *c_p;
    c_p++;
  }
}

void MultiplyVect(
  const std::vector<int>& in1,
  const std::vector<int>& in2,
  std::vector<int>* out)
{
  out->resize(0);
  std::vector<int>::const_iterator in1_p, in2_p;
  for(in1_p=in1.begin(), in2_p=in2.begin(); in1_p!=in1.end(); in1_p++, in2_p++)
    out->push_back((*in1_p)*(*in2_p));
}

void AddVect(
  const std::vector<int>& in1,
  const std::vector<int>& in2,
  std::vector<int>* out)
{
  out->resize(0);
  std::vector<int>::const_iterator in1_p, in2_p;
  for(in1_p=in1.begin(), in2_p=in2.begin(); in1_p!=in1.end(); in1_p++, in2_p++)
    out->push_back((*in1_p) + (*in2_p));
}

void OperateVect(
  const std::vector<int>& in1,
  const std::vector<int>& in2,
  const std::vector<int>& in3,
  const std::vector<int>& in4,
  std::vector<int>* out)
{
  out->resize(0);
  std::vector<int> tmp1, tmp2;
  MultiplyVect(in1,in2,&tmp1);
  MultiplyVect(in3,in4,&tmp2);
  AddVect(tmp1,tmp2,out);
}

};


void DiffuseRegions(
  const ImgInt& in, 
  const int area_limit, 
  ImgInt* out,
  ImgInt* out_original, 
  std::vector<int>* previous_labels)
{
  assert(Min(in) >= 0);

  ImgInt in_reorder;
  ConnectedComponents(in,&in_reorder);

  std::vector<int> reorder_index;
  std::vector<int> reorder_npixels;
  std::vector<Point> reorder_first_point;
  std::vector<int> in_values;
  ConnectedComponentsInfo(in,in_reorder,&reorder_index,&reorder_npixels,
                          &reorder_first_point,&in_values);

  ImgInt::Iterator inr_p;
  for(inr_p=in_reorder.Begin(); inr_p!=in_reorder.End(); inr_p++)
  {
    if(reorder_npixels[*inr_p] > area_limit)
    {
      reorder_index[*inr_p] = (*inr_p) + 1;
      *inr_p = (*inr_p) + 1;
    }     
    else
    {
      reorder_index[*inr_p] = 0;
      *inr_p = 0;
    }
  }

  ImgInt shed;
  ConnectedComponents(in_reorder,&shed);

  std::vector<int> shed_index;
  std::vector<int> shed_npixels;
  std::vector<Point> shed_first_point;
  std::vector<int> in_reorder_values;
  ConnectedComponentsInfo(in_reorder,shed,&shed_index,&shed_npixels,
                          &shed_first_point,&in_reorder_values);

  std::vector<int> in_shed;
  std::vector<int> shed_replace;
  std::vector<int> small_region_values;
  std::vector<Point> small_region_first_point;
  std::vector<int>::const_iterator shnp_p, shi_p;
  std::vector<Point>::const_iterator shfp_p;
  shnp_p = shed_npixels.begin();
  shi_p = shed_index.begin();
  shfp_p = shed_first_point.begin();
  while(shnp_p != shed_npixels.end())
  {
    if(in_reorder_values[*shi_p] == 0)
    // ensures small regions combining into big regions are also removed
    {
      small_region_values.push_back(*shi_p);
      small_region_first_point.push_back(*shfp_p);
      shed_replace.push_back(-1);
      in_shed.push_back(-1);
    }
    else
    {
      shed_replace.push_back(*shi_p);
      in_shed.push_back(in_values[(in_reorder_values[*shi_p])-1]);
    }
    shnp_p++;
    shi_p++;
    shfp_p++;
  }
  
  std::vector<PointArray> boundaries;
  std::vector<intArray> boundary_values;
  std::vector<bool> unique_boundary;
  RegionBoundaries(shed,small_region_values,small_region_first_point,
                   &boundaries,&boundary_values,&unique_boundary);

  std::vector<int> replace_values;
  ValueToReplace(shed,small_region_values,boundary_values,unique_boundary,&replace_values);

  std::vector<int>::const_iterator srv_p, rv_p;
  srv_p = small_region_values.begin();
  rv_p = replace_values.begin();
  while(srv_p != small_region_values.end())
  {
    shed_replace[*srv_p] = *rv_p;
    srv_p++;
    rv_p++;
  }

  ImgInt out_unordered;
  out_unordered.Reset(shed.Width(), shed.Height());
  ImgInt::Iterator outu_p = out_unordered.Begin();
  ImgInt::ConstIterator shed_p;
  for(shed_p=shed.Begin(); shed_p!=shed.End(); shed_p++)
  {
    *outu_p = shed_replace[*shed_p];
    outu_p++;
  }

  ConnectedComponents(out_unordered,out);

  std::vector<int> out_index;
  std::vector<int> out_npixels;
  std::vector<Point> out_first_point;
  std::vector<int> out_shed_values;
  ConnectedComponentsInfo(out_unordered,*out,&out_index,&out_npixels,
                          &out_first_point,&out_shed_values);

  std::vector<int>::const_iterator oi_p, osv_p;
  oi_p = out_index.begin();
  osv_p = out_shed_values.begin();
  while(oi_p != out_index.end())
  {
    previous_labels->push_back(in_shed[*osv_p]);
    oi_p++;
    osv_p++;
  }

  out_original->Reset(out->Width(), out->Height());
  ImgInt::Iterator outo_p, out_p;
  out_p = out->Begin();
  for(outo_p=out_original->Begin(); outo_p!=out_original->End(); outo_p++)
  {
    *outo_p = (*previous_labels)[*out_p];
    out_p++;
  }
}

void CoerceRegions(
  const ImgInt& in, 
  const int area_limit, 
  ImgInt* out,
  ImgInt* out_original, 
  std::vector<int>* previous_labels)
{
  assert(Min(in) >= 0);

  int width = in.Width();
  int height = in.Height();

  ImgInt in_reorder;
  ConnectedComponents(in,&in_reorder);

  std::vector<int> reorder_index;
  std::vector<int> reorder_npixels;
  std::vector<Point> reorder_first_point;
  std::vector<int> in_values;
  ConnectedComponentsInfo(in,in_reorder,&reorder_index,&reorder_npixels,
                          &reorder_first_point,&in_values);

  ImgInt::Iterator inr_p;
  for(inr_p=in_reorder.Begin(); inr_p!=in_reorder.End(); inr_p++)
  {
    if(reorder_npixels[*inr_p] > area_limit)
    {
      reorder_index[*inr_p] = (*inr_p) + 1;
      *inr_p = (*inr_p) + 1;
    }     
    else
    {
      reorder_index[*inr_p] = 0;
      *inr_p = 0;
    }
  }

  *out_original = in_reorder;

  int i,j;
  for(j=(height-2); j>=0; j--)
    for(i=0; i<width; i++)
      if((*out_original)(i,j) == 0)
        (*out_original)(i,j) = (*out_original)(i,j+1);
  for(j=2; j<height; j++)
    for(i=0; i<width; i++)
      if((*out_original)(i,j) == 0)
        (*out_original)(i,j) = (*out_original)(i,j-1);
  for(i=2; i<width; i++)
    for(j=0; j<height; j++)
      if((*out_original)(i,j) == 0)
        (*out_original)(i,j) = (*out_original)(i-1,j);
  for(i=(width-2); i>=0; i--)
    for(j=0; j<height; j++)
      if((*out_original)(i,j) == 0)
        (*out_original)(i,j) = (*out_original)(i+1,j);

  /*std::vector<int> row, row_out_cur, row_out_pre, mask_pos, mask_neg;
  for(int j=2; j<height; j++)
  {
    ExtractRow(in_reorder,j,&row);
    ExtractRow(*out_original,j-1,&row_out_pre);
    ExtractRowZero(row,&mask_pos,&mask_neg);
    OperateVect(row,mask_neg,row_out_pre,mask_pos,&row_out_cur);
    SetRow(row_out_cur,j,*out_original);
  }
  for(j=(height-2); j>=0; j--)
  {
    ExtractRow(*out_original,j,&row);
    ExtractRow(*out_original,j+1,&row_out_pre);
    ExtractRowZero(row,&mask_pos,&mask_neg);
    OperateVect(row,mask_neg,row_out_pre,mask_pos,&row_out_cur);
    SetRow(row_out_cur,j,*out_original);
  }

  for(int i=2; i<width; i++)
  {
    ExtractCol(*out_original,i,&row);
    ExtractCol(*out_original,i-1,&row_out_pre);
    ExtractRowZero(row,&mask_pos,&mask_neg);
    OperateVect(row,mask_neg,row_out_pre,mask_pos,&row_out_cur);
    SetCol(row_out_cur,i,*out_original);
  }
  for(i=(width-2); i>=0; i--)
  {
    ExtractCol(*out_original,i,&row);
    ExtractCol(*out_original,i+1,&row_out_pre);
    ExtractRowZero(row,&mask_pos,&mask_neg);
    OperateVect(row,mask_neg,row_out_pre,mask_pos,&row_out_cur);
    SetCol(row_out_cur,i,*out_original);
  }*/

  ConnectedComponents(*out_original, out);

  std::vector<int> out_index, out_npixels, out_orig_labels;
  std::vector<Point> out_first_point;
  ConnectedComponentsInfo(*out_original,*out,&out_index,&out_npixels,
                          &out_first_point,&out_orig_labels);

  std::vector<int>::const_iterator oi_p, ol_p;
  oi_p = out_index.begin();
  ol_p = out_orig_labels.begin();
  while(oi_p != out_index.end())
  {
    previous_labels->push_back(in_values[(*ol_p)-1]);
    oi_p++;
    ol_p++;
  }

  /*Figure fig;
  fig.Draw(*out);
  AfxMessageBox("ok");*/
  
}


